/*
 * This file is a part of the CaosDB Project.
 *
 * Copyright (C) 2021 Timm Fitschen <t.fitschen@indiscale.com>
 * Copyright (C) 2021 IndiScale GmbH <info@indiscale.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 */

#ifndef CAOSDB_VALUE_H
#define CAOSDB_VALUE_H
#include "caosdb/protobuf_helper.h"         // for ProtoMessageWrapper
#include "caosdb/entity/v1alpha1/main.pb.h" // for RepeatedPtrField, Message

#include <google/protobuf/util/json_util.h> // for MessageToJson...
#include <memory>                           // for unique_ptr
#include <string>                           // for string
#include <vector>                           // for vector

#define LIST_VALUE_CONSTRUCTOR(TYPE, SETTER)                                   \
  explicit inline Value(const std::vector<TYPE> &values)                       \
    : ProtoMessageWrapper<ProtoValue>() {                                      \
    for (const auto &value : values) {                                         \
      this->wrapped->mutable_list_values()->add_values()->SETTER(value);       \
    }                                                                          \
  }

namespace caosdb::entity {
using caosdb::utility::ProtoMessageWrapper;
using ProtoSpecialValue = caosdb::entity::v1alpha1::SpecialValue;
using ProtoValue = caosdb::entity::v1alpha1::Value;
using ProtoScalarValue = caosdb::entity::v1alpha1::ScalarValue;
using ValueCase = caosdb::entity::v1alpha1::Value::ValueCase;
using ScalarValueCase = caosdb::entity::v1alpha1::ScalarValue::ScalarValueCase;

class Entity;
class Property;

// Represents special values which are otherwise hard to tranfer via protobuf.
enum SpecialValue {
  // Represent the NULL value.
  NULL_VALUE = ProtoSpecialValue::SPECIAL_VALUE_UNSPECIFIED,
  // The empty string.
  EMPTY_STRING = ProtoSpecialValue::SPECIAL_VALUE_EMPTY_STRING,
};

class ScalarValue : public ProtoMessageWrapper<ProtoScalarValue> {
public:
  inline ScalarValue(ProtoScalarValue *wrapped)
    : ProtoMessageWrapper<ProtoScalarValue>(wrapped) {}

  [[nodiscard]] inline auto IsString() const noexcept -> bool {
    return (this->wrapped->scalar_value_case() ==
            ScalarValueCase::kStringValue) ||
           (this->wrapped->scalar_value_case() ==
              ScalarValueCase::kSpecialValue &&
            this->wrapped->special_value() ==
              ProtoSpecialValue::SPECIAL_VALUE_EMPTY_STRING);
    return false;
  }
  [[nodiscard]] inline auto AsString() const noexcept -> const std::string & {
    return this->wrapped->string_value();
    ;
  }

  [[nodiscard]] inline auto IsDouble() const noexcept -> bool {
    return (this->wrapped->scalar_value_case() ==
            ScalarValueCase::kDoubleValue);
  }
  [[nodiscard]] inline auto AsDouble() const noexcept -> double {
    return this->wrapped->double_value();
  }

  [[nodiscard]] inline auto IsInteger() const noexcept -> bool {
    return (this->wrapped->scalar_value_case() ==
            ScalarValueCase::kIntegerValue);
  }
  [[nodiscard]] inline auto AsInteger() const noexcept -> int64_t {
    return this->wrapped->integer_value();
  }

  [[nodiscard]] inline auto IsBool() const noexcept -> bool {
    return (this->wrapped->scalar_value_case() ==
            ScalarValueCase::kBooleanValue);
  }
  [[nodiscard]] inline auto AsBool() const noexcept -> bool {
    return this->wrapped->boolean_value();
  }
};

class Value : public ProtoMessageWrapper<ProtoValue> {
public:
  inline Value() : ProtoMessageWrapper<ProtoValue>() {
    // has NULL_VALUE now
  }
  explicit inline Value(const std::string &value)
    : ProtoMessageWrapper<ProtoValue>() {
    this->wrapped->mutable_scalar_value()->set_string_value(value);
  }
  explicit inline Value(const char value[]) : Value(std::string(value)) {}
  explicit inline Value(double value) : ProtoMessageWrapper<ProtoValue>() {
    this->wrapped->mutable_scalar_value()->set_double_value(value);
  }
  explicit inline Value(int64_t value) : ProtoMessageWrapper<ProtoValue>() {
    this->wrapped->mutable_scalar_value()->set_integer_value(value);
  }
  explicit inline Value(int value) : Value((int64_t)value) {}
  explicit inline Value(bool value) : ProtoMessageWrapper<ProtoValue>() {
    this->wrapped->mutable_scalar_value()->set_boolean_value(value);
  }

  explicit inline Value(const std::vector<std::string> &values)
    : ProtoMessageWrapper<ProtoValue>() {
    for (const auto &value : values) {
      this->wrapped->mutable_list_values()->add_values()->set_string_value(
        value);
    }
  }

  LIST_VALUE_CONSTRUCTOR(int64_t, set_integer_value)

  [[nodiscard]] inline auto IsNull() -> bool {
    return this->wrapped->value_case() == ValueCase::VALUE_NOT_SET;
  }

  [[nodiscard]] inline auto IsString() const noexcept -> bool {
    if (this->wrapped->value_case() == ValueCase::kScalarValue) {

      return (this->wrapped->scalar_value().scalar_value_case() ==
              ScalarValueCase::kStringValue) ||
             (this->wrapped->scalar_value().scalar_value_case() ==
                ScalarValueCase::kSpecialValue &&
              this->wrapped->scalar_value().special_value() ==
                ProtoSpecialValue::SPECIAL_VALUE_EMPTY_STRING);
    }
    return false;
  }
  [[nodiscard]] inline auto AsString() const noexcept -> const std::string & {
    return this->wrapped->scalar_value().string_value();
    ;
  }

  [[nodiscard]] inline auto IsDouble() const noexcept -> bool {
    if (this->wrapped->value_case() == ValueCase::kScalarValue) {

      return (this->wrapped->scalar_value().scalar_value_case() ==
              ScalarValueCase::kDoubleValue);
    }
    return false;
  }
  [[nodiscard]] inline auto AsDouble() const noexcept -> double {
    return this->wrapped->scalar_value().double_value();
  }

  [[nodiscard]] inline auto IsInteger() const noexcept -> bool {
    if (this->wrapped->value_case() == ValueCase::kScalarValue) {

      return (this->wrapped->scalar_value().scalar_value_case() ==
              ScalarValueCase::kIntegerValue);
    }
    return false;
  }
  [[nodiscard]] inline auto AsInteger() const noexcept -> int64_t {
    return this->wrapped->scalar_value().integer_value();
  }

  [[nodiscard]] inline auto IsBool() const noexcept -> bool {
    if (this->wrapped->value_case() == ValueCase::kScalarValue) {

      return (this->wrapped->scalar_value().scalar_value_case() ==
              ScalarValueCase::kBooleanValue);
    }
    return false;
  }
  [[nodiscard]] inline auto AsBool() const noexcept -> bool {
    return this->wrapped->scalar_value().boolean_value();
  }

  [[nodiscard]] inline auto IsList() const noexcept -> bool {
    return this->wrapped->value_case() == ValueCase::kListValues;
  }
  [[nodiscard]] inline auto AsList() const noexcept
    -> const std::vector<ScalarValue> & {
    if (!IsList()) {
      // create empty list
      this->list_values = std::make_unique<std::vector<ScalarValue>>();
    }
    if (this->list_values == nullptr) {
      this->list_values = std::make_unique<std::vector<ScalarValue>>();
      for (auto &scalar :
           *(this->wrapped->mutable_list_values()->mutable_values())) {
        this->list_values->push_back(ScalarValue(&scalar));
      }
    }
    return *(this->list_values);
  }

  inline auto operator==(const Value &other) const noexcept -> bool {
    return this->wrapped->SerializeAsString() ==
           other.wrapped->SerializeAsString();
  }

  inline auto ToString() const noexcept -> const std::string {
    CAOSDB_DEBUG_MESSAGE_STRING(*wrapped, out)
    return out;
  }

  friend class Entity;
  friend class Property;

private:
  mutable std::unique_ptr<std::vector<ScalarValue>> list_values;
};

} // namespace caosdb::entity

#endif
